@@ -1,23 +1,25 @@
 ;;; emms-info.el --- Retrieving track information
 
-;; Copyright (C) 2005, 2006 Free Software Foundation Inc.
+;; Copyright (C) 2005, 2006, 2007 Free Software Foundation Inc.
 
 ;; Author: Jorgen Schaefer <forcer@forcix.cx>
 
-;; This program is free software; you can redistribute it and/or
+;; This file is part of EMMS.
+
+;; EMMS is free software; you can redistribute it and/or
 ;; modify it under the terms of the GNU General Public License
-;; as published by the Free Software Foundation; either version 2
+;; as published by the Free Software Foundation; either version 3
 ;; of the License, or (at your option) any later version.
 
-;; This program is distributed in the hope that it will be useful,
+;; EMMS is distributed in the hope that it will be useful,
 ;; but WITHOUT ANY WARRANTY; without even the implied warranty of
 ;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with this program; if not, write to the Free Software
+;; along with EMMS; if not, write to the Free Software
 ;; Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
-;; 02110-1301  USA
+;; 02110-1301, USA.
 
 ;;; Commentary:
 
@@ -58,6 +60,13 @@ This requires `later-do', which should come with EMMS."
   :type 'boolean
   :group 'emms-info)
 
+(defcustom emms-info-report-each-num-tracks 200
+  "*Non-zero will report progress information every number of tracks.
+The default is to display a message every 200 tracks.
+This variable is only used when adding tracks asynchronously."
+  :type 'integer
+  :group 'emms-info)
+
 (defcustom emms-info-functions nil
   "*Functions which add information to tracks.
 Each is called with a track as argument."
@@ -78,30 +87,39 @@ This is a suitable value for `emms-track-initialize-functions'."
 (defun emms-info-really-initialize-track (track)
   "Really initialize TRACK.
 Return t when the track got changed."
-  (let ((track-mtime (emms-track-get track 'info-mtime))
-        (file-mtime (when emms-info-auto-update
-                      (emms-info-track-file-mtime track))))
-    (when (or (not track-mtime)
-              (when emms-info-auto-update
-                (emms-time-less-p track-mtime
-                                  file-mtime)))
-      (run-hook-with-args 'emms-info-functions
-                          track)
-      (emms-track-set track 'info-mtime file-mtime)
-      (emms-track-updated track)
-      (when emms-info-asynchronously
-        (setq emms-info-asynchronous-tracks (1- emms-info-asynchronous-tracks))
-        (when (zerop emms-info-asynchronous-tracks)
-          (message "EMMS: All track information loaded.")))
-      t)))
+  (let ((file-mtime (when emms-info-auto-update
+                      (emms-info-track-file-mtime track)))
+        (info-mtime (emms-track-get track 'info-mtime))
+        (name (emms-track-get track 'name)))
+
+    ;; if the file's been modified or is new
+    (when (or (not file-mtime)
+              (not info-mtime)
+              (emms-time-less-p info-mtime file-mtime))
+      (run-hook-with-args 'emms-info-functions track)
+      ;; not set by info functions
+      (when file-mtime
+        (emms-track-set track 'info-mtime file-mtime))
+      (emms-track-updated track))
+
+    (when emms-info-asynchronously
+      (setq emms-info-asynchronous-tracks (1- emms-info-asynchronous-tracks))
+      (if (zerop emms-info-asynchronous-tracks)
+          (message "EMMS: All track information loaded.")
+        (unless (zerop emms-info-report-each-num-tracks)
+          (if (zerop
+               (mod emms-info-asynchronous-tracks
+                    emms-info-report-each-num-tracks))
+              (message "EMMS: %d tracks to go.."
+                       emms-info-asynchronous-tracks)))))))
 
 (defun emms-info-track-file-mtime (track)
   "Return the mtime of the file of TRACK, if any.
-Return zero otherwise."
+Return nil otherwise."
   (if (eq (emms-track-type track)
           'file)
       (nth 5 (file-attributes (emms-track-name track)))
-    0))
+    nil))
 
 (defun emms-info-track-description (track)
   "Return a description of the current track."
