@@ -1,6 +1,6 @@
 ;;; emms-player-mpd.el --- MusicPD support for EMMS
 
-;; Copyright (C) 2005, 2006 Free Software Foundation, Inc.
+;; Copyright (C) 2005, 2006, 2007 Free Software Foundation, Inc.
 
 ;; Author: Michael Olson <mwolson@gnu.org>
 
@@ -8,7 +8,7 @@
 
 ;; EMMS is free software; you can redistribute it and/or modify it
 ;; under the terms of the GNU General Public License as published by
-;; the Free Software Foundation; either version 2, or (at your option)
+;; the Free Software Foundation; either version 3, or (at your option)
 ;; any later version.
 ;;
 ;; EMMS is distributed in the hope that it will be useful, but WITHOUT
@@ -36,9 +36,7 @@
 ;; installation is needed.
 
 ;; The website is at http://musicpd.org/.  Debian packages are
-;; available.  I recommend getting the latest development version; see
-;; http://mpd.wikicities.com/wiki/Subversion for nightly Debian
-;; packages and the svn repo.
+;; available.  It is recommended to use mpd version 0.12.0 or higher.
 ;;
 ;; Copy the example configuration for mpd into ~/.mpdconf and edit it
 ;; to your needs.  Use your top level music directory for
@@ -64,21 +62,25 @@
 ;; `emms-player-mpd-server-port' to match the location and port of
 ;; your MusicPD server.
 ;;
-;; (setq emms-player-mpd-server-name "localhost")
-;; (setq emms-player-mpd-server-port "6600")
+;;   (setq emms-player-mpd-server-name "localhost")
+;;   (setq emms-player-mpd-server-port "6600")
 
 ;; If your MusicPD setup requires a password, you will need to do the
 ;; following.
 ;;
-;; (setq emms-player-mpd-server-password "mypassword")
+;;   (setq emms-player-mpd-server-password "mypassword")
 
 ;; To get track info from MusicPD, do the following.
 ;;
-;; (add-to-list 'emms-info-functions 'emms-info-mpd)
+;;   (add-to-list 'emms-info-functions 'emms-info-mpd)
+
+;; To change the volume using MusicPD, do the following.
+;;
+;;   (setq emms-volume-change-function 'emms-volume-mpd-change)
 
 ;; Add 'emms-player-mpd to the top of `emms-player-list'.
 ;;
-;; (add-to-list 'emms-player-list 'emms-player-mpd)
+;;   (add-to-list 'emms-player-list 'emms-player-mpd)
 
 ;; If you use absolute file names in your m3u playlists (which is most
 ;; likely), make sure you set `emms-player-mpd-music-directory' to the
@@ -104,6 +106,13 @@
 (require 'emms-source-playlist)  ; for emms-source-file-parse-playlist
 (require 'tq)
 
+(eval-when-compile
+  (condition-case nil
+      (progn
+        (require 'url)           ; load if available
+        (require 'emms-url))
+    (error nil)))
+
 (defgroup emms-player-mpd nil
   "EMMS player for MusicPD."
   :group 'emms-player
@@ -118,6 +127,7 @@
 
 (defcustom emms-player-mpd-music-directory nil
   "The value of 'music_directory' in your MusicPD configuration file.
+
 You need this if your playlists use absolute file names, otherwise
 leave it set to nil."
   ;; The :format part ensures that entering directories happens on the
@@ -138,8 +148,9 @@ or nil if we cannot figure it out."
     (while (car out)
       (cond ((string= (car out) "Supported formats:")
              (setq found-start t))
-            ((and found-start
-                  (not (string= (car out) "")))
+            ((string= (car out) "")
+             (setq found-start nil))
+            (found-start
              (setq supported (concat supported (car out)))))
       (setq out (cdr out)))
     ;; Create regexp
@@ -154,7 +165,7 @@ or nil if we cannot figure it out."
   (or (emms-player-mpd-get-supported-regexp)
       (concat "\\`http://\\|"
               "\\.\\(m3u\\|ogg\\|flac\\|mp3\\|wav\\|mod\\|au\\|aiff\\)\\'"))
-  "Formats supported by MusicPD Client."
+  "Formats supported by MusicPD."
   :type 'regexp
   :set (function
         (lambda (sym value)
@@ -179,7 +190,7 @@ It should take same arguments as `open-network-stream' does."
   :group 'emms-player-mpd)
 
 (defcustom emms-player-mpd-server-password nil
-  "The port of the MusicPD server that we should connect to."
+  "The password for the MusicPD server that we should connect to."
   :type '(choice (const :tag "None" nil)
                  string)
   :group 'emms-player-mpd)
@@ -189,7 +200,7 @@ It should take same arguments as `open-network-stream' does."
 next song.  This may be an integer or a floating point number.
 
 This is used only if `emms-player-mpd-sync-playlist' is non-nil"
-  :type 'integer
+  :type 'number
   :group 'emms-player-mpd)
 
 (defcustom emms-player-mpd-verbose nil
@@ -199,7 +210,7 @@ and errors."
   :group 'emms-player-mpd)
 
 (defcustom emms-player-mpd-sync-playlist t
-  "Whether to syncronize the EMMS playlist with the MusicPD playlist.
+  "Whether to synchronize the EMMS playlist with the MusicPD playlist.
 
 If your EMMS playlist contains music files rather than playlists,
 leave this set to non-nil.
@@ -224,6 +235,10 @@ If your EMMS playlist contains stored playlists, set this to nil."
                  'seek
                  'emms-player-mpd-seek)
 
+(emms-player-set emms-player-mpd
+                 'seek-to
+                 'emms-player-mpd-seek-to)
+
 ;;; Dealing with the MusicPD network process
 
 (defvar emms-player-mpd-process nil)
@@ -309,6 +324,7 @@ When a reply comes, call FN with CLOSURE and the result."
 
 (defun emms-player-mpd-get-mpd-filename (file)
   "Turn FILE into something that MusicPD can understand.
+
 This usually means removing a prefix."
   (if (or (not emms-player-mpd-music-directory)
           (not (eq (aref file 0) ?/))
@@ -318,6 +334,7 @@ This usually means removing a prefix."
 
 (defun emms-player-mpd-get-emms-filename (file)
   "Turn FILE into something that EMMS can understand.
+
 This usually means adding a prefix."
   (if (or (not emms-player-mpd-music-directory)
           (eq (aref file 0) ?/)
@@ -327,6 +344,7 @@ This usually means adding a prefix."
 
 (defun emms-player-mpd-parse-response (response)
   "Convert the given MusicPD response into a list.
+
 The car of the list is special:
 If an error has occurred, it will contain a cons cell whose car is
 an error number and whose cdr is the corresponding message.
@@ -352,6 +370,7 @@ Otherwise, it will be nil."
 
 (defun emms-player-mpd-parse-line (line)
   "Turn the given LINE from MusicPD into a cons cell.
+
 The format of the cell is (name . value)."
   (when (string-match "\\`\\([^:\n]+\\):\\s-*\\(.+\\)" line)
     (let ((name (match-string 1 line))
@@ -363,8 +382,7 @@ The format of the cell is (name . value)."
         nil))))
 
 (defun emms-player-mpd-get-alist (info)
-  "Turn the given parsed INFO from MusicPD into an alist.
-The format of the alist is (name . value)."
+  "Turn the given parsed INFO from MusicPD into an alist."
   (when (and info
              (null (car info))          ; no error has occurred
              (cdr info))                ; data exists
@@ -379,7 +397,6 @@ The format of the alist is (name . value)."
 
 (defun emms-player-mpd-get-alists (info)
   "Turn the given parsed INFO from MusicPD into an list of alists.
-The format of the alist is (name . value).
 
 The list will be in reverse order."
   (when (and info
@@ -517,6 +534,44 @@ info from MusicPD."
                      (string-to-number (match-string 1 time)))))
      "time" info)))
 
+(defun emms-player-mpd-select-song (prev-song new-song)
+  "Move to the given song position.
+
+The amount to move is the number difference between PREV-SONG and
+NEW-SONG.  NEW-SONG should be a string containing a number.
+PREV-SONG may be either a string containing a number or nil,
+which indicates that we should start from the beginning of the
+buffer and move to NEW-SONG."
+  (with-current-emms-playlist
+    ;; move to current track
+    (goto-char (if (and (stringp prev-song)
+                        emms-playlist-selected-marker
+                        (marker-position emms-playlist-selected-marker))
+                   emms-playlist-selected-marker
+                 (point-min)))
+    ;; seek forward or backward
+    (let ((diff (if (stringp prev-song)
+                    (- (string-to-number new-song)
+                       (string-to-number prev-song))
+                  (string-to-number new-song))))
+      (condition-case nil
+          (progn
+            ;; skip to first track if not on one
+            (when (and (> diff 0)
+                       (not (emms-playlist-track-at (point))))
+              (emms-playlist-next))
+            ;; move to new track
+            (while (> diff 0)
+              (emms-playlist-next)
+              (setq diff (- diff 1)))
+            (while (< diff 0)
+              (emms-playlist-previous)
+              (setq diff (+ diff 1)))
+            ;; select track at point
+            (unless (emms-playlist-selected-track-at-p)
+              (emms-playlist-select (point))))
+        (error (concat "Could not move to position " new-song))))))
+
 (defun emms-player-mpd-sync-from-emms-1 (closure)
   (emms-player-mpd-get-playlist-id
    closure
@@ -556,9 +611,7 @@ errors."
           (setq emms-player-mpd-playlist-id id)
           (set-buffer-modified-p nil)
           (if song
-              (progn
-                (goto-line (1+ (string-to-number song)))
-                (emms-playlist-select (point)))
+              (emms-player-mpd-select-song nil song)
             (goto-char (point-min)))))
       (when (functionp fn)
         (funcall fn close info)))))
@@ -585,33 +638,38 @@ MusicPD playlist."
 (defun emms-player-mpd-detect-song-change-1 (closure info)
   (let ((song (emms-player-mpd-get-current-song nil #'ignore info))
         (status (emms-player-mpd-get-mpd-state nil #'ignore info))
-        (time (emms-player-mpd-get-playing-time nil #'ignore info)))
-    (cond ((string= status "stop")
-           (emms-player-mpd-disconnect t)
-           (if song
-               ;; a track remains: the user probably stopped MusicPD
-               ;; manually, so we'll stop EMMS completely
+        (time (emms-player-mpd-get-playing-time nil #'ignore info))
+        (err-msg (cdr (assoc "error" info))))
+    (if (stringp err-msg)
+        (progn
+          (message "MusicPD error: %s" err-msg)
+          (emms-player-mpd-send
+           "clearerror"
+           nil #'ignore))
+      (cond ((string= status "stop")
+             (emms-player-mpd-disconnect t)
+             (if song
+                 ;; a track remains: the user probably stopped MusicPD
+                 ;; manually, so we'll stop EMMS completely
+                 (let ((emms-player-stopped-p t))
+                   (emms-player-stopped))
+               ;; no more tracks are left: we probably ran out of things
+               ;; to play, so let EMMS do something further if it wants
+               (emms-player-stopped)))
+            ((string= status "pause")
+             nil)
+            ((string= status "play")
+             (unless (or (null song)
+                         (and (stringp emms-player-mpd-current-song)
+                              (string= song emms-player-mpd-current-song)))
                (let ((emms-player-stopped-p t))
                  (emms-player-stopped))
-             ;; no more tracks are left: we probably ran out of things
-             ;; to play, so let EMMS do something further if it wants
-             (emms-player-stopped)))
-          ((string= status "pause")
-           nil)
-          ((string= status "play")
-           (unless (or (null song)
-                       (and (stringp emms-player-mpd-current-song)
-                            (string= song emms-player-mpd-current-song)))
-             (setq emms-player-mpd-current-song song)
-             (let ((emms-player-stopped-p t))
-               (emms-player-stopped))
-             (with-current-emms-playlist
-               (emms-playlist-select (progn
-                                       (goto-line (1+ (string-to-number song)))
-                                       (point))))
-             (emms-player-started 'emms-player-mpd)
-             (when time
-               (run-hook-with-args 'emms-player-time-set-functions time)))))))
+               (emms-player-mpd-select-song emms-player-mpd-current-song song)
+               (setq emms-player-mpd-current-song song)
+               (emms-player-started 'emms-player-mpd)
+               (when time
+                 (run-hook-with-args 'emms-player-time-set-functions
+                                     time))))))))
 
 (defun emms-player-mpd-detect-song-change (&optional info)
   "Detect whether a song change has occurred.
@@ -642,9 +700,9 @@ info from MusicPD."
 
 (defun emms-player-mpd-add-file (file closure callback)
   "Add FILE to the current MusicPD playlist.
-If an error occurs, display a relevant message.
+Execute CALLBACK with CLOSURE as its first argument when done.
 
-Execute CALLBACK with CLOSURE as its first argument when done."
+If an error occurs, display a relevant message."
   (setq file (emms-player-mpd-get-mpd-filename file))
   (emms-player-mpd-send
    (concat "add " (emms-player-mpd-quote-file file))
@@ -659,45 +717,46 @@ Execute CALLBACK with CLOSURE as its first argument when done."
          (when (functionp callback)
            (funcall callback close)))))))
 
-(defun emms-player-mpd-add-buffer-contents (closure callback)
-  "Load contents of the current buffer into MusicPD by adding each line.
-This handles both m3u and pls type playlists.
+(defun emms-player-mpd-add-buffer-contents (buffer closure callback)
+  "Load contents of BUFFER into MusicPD by adding each line.
+Execute CALLBACK with CLOSURE as its first argument when done.
 
-Execute CALLBACK with CLOSURE as its first argument when done."
-  (goto-char (point-min))
-  (let ((format (emms-source-playlist-determine-format)))
-    (when format
-      (emms-player-mpd-add-several-files
-       (emms-source-playlist-files format)
-       closure callback))))
+This handles both m3u and pls type playlists."
+  (with-current-buffer buffer
+    (goto-char (point-min))
+    (let ((format (emms-source-playlist-determine-format)))
+      (when format
+        (emms-player-mpd-add-several-files
+         (emms-source-playlist-files format)
+         closure callback)))))
 
 (defun emms-player-mpd-add-playlist (playlist closure callback)
   "Load contents of PLAYLIST into MusicPD by adding each line.
-This handles both m3u and pls type playlists.
+Execute CALLBACK with CLOSURE as its first argument when done.
 
-Execute CALLBACK with CLOSURE as its first argument when done."
+This handles both m3u and pls type playlists."
   ;; This is useful for playlists of playlists
   (with-temp-buffer
     (insert-file-contents playlist)
-    (emms-player-mpd-add-buffer-contents closure callback)))
+    (emms-player-mpd-add-buffer-contents (current-buffer) closure callback)))
 
 (defun emms-player-mpd-add-streamlist (url closure callback)
   "Download contents of URL and then add its feeds into MusicPD.
-
 Execute CALLBACK with CLOSURE as its first argument when done."
   ;; This is useful with emms-streams.el
-  (condition-case nil
+  (if (fboundp 'url-insert-file-contents)
       (progn
-        (require 'url)
+        (require 'emms-url)
         (with-temp-buffer
-          (url-insert-file-contents url)
-          (emms-player-mpd-add-buffer-contents closure callback)))
+          (url-insert-file-contents (emms-escape-url url))
+          (emms-http-decode-buffer (current-buffer))
+          (emms-player-mpd-add-buffer-contents (current-buffer)
+                                               closure callback)))
     (error (message (concat "You need to install url.el so that"
                             " Emms can retrieve this stream")))))
 
 (defun emms-player-mpd-add (track closure callback)
   "Add TRACK to the MusicPD playlist.
-
 Execute CALLBACK with CLOSURE as its first argument when done."
   (let ((name (emms-track-get track 'name))
         (type (emms-track-get track 'type)))
@@ -713,7 +772,6 @@ Execute CALLBACK with CLOSURE as its first argument when done."
 
 (defun emms-player-mpd-add-several-tracks (tracks closure callback)
   "Add TRACKS to the MusicPD playlist.
-
 Execute CALLBACK with CLOSURE as its first argument when done."
   (when (consp tracks)
     (while (cdr tracks)
@@ -724,7 +782,6 @@ Execute CALLBACK with CLOSURE as its first argument when done."
 
 (defun emms-player-mpd-add-several-files (files closure callback)
   "Add FILES to the MusicPD playlist.
-
 Execute CALLBACK with CLOSURE as its first argument when done."
   (when (consp files)
     (while (cdr files)
@@ -768,33 +825,51 @@ playlist."
      (lambda (closure response)
        (emms-player-started 'emms-player-mpd)))))
 
-(defun emms-player-mpd-start-and-sync-1 (buffer id)
-  (when emms-player-mpd-status-timer
-    (emms-cancel-timer emms-player-mpd-status-timer)
-    (setq emms-player-mpd-status-timer nil))
+(defun emms-player-mpd-start-and-sync-2 (buffer id)
   (when (buffer-live-p buffer)
     (let ((emms-playlist-buffer buffer))
       (with-current-emms-playlist
         (setq emms-player-mpd-playlist-id id)
         (set-buffer-modified-p nil)
-        (emms-player-mpd-play (1- (emms-line-number-at-pos
-                                   emms-playlist-selected-marker)))))))
+        (let ((track-cnt 0))
+          (save-excursion
+            (goto-char
+             (if (and emms-playlist-selected-marker
+                      (marker-position emms-playlist-selected-marker))
+                 emms-playlist-selected-marker
+               (point-min)))
+            (condition-case nil
+                (while t
+                  (emms-playlist-previous)
+                  (setq track-cnt (1+ track-cnt)))
+              (error nil)))
+          (emms-player-mpd-play track-cnt))))))
+
+(defun emms-player-mpd-start-and-sync-1 (closure id)
+  (let ((buf-id (with-current-emms-playlist
+                  emms-player-mpd-playlist-id)))
+    (if (and (not (buffer-modified-p emms-playlist-buffer))
+             (stringp buf-id)
+             (string= buf-id id))
+        (emms-player-mpd-start-and-sync-2 emms-playlist-buffer id)
+      (emms-player-mpd-sync-from-emms
+       #'emms-player-mpd-start-and-sync-2))))
 
 (defun emms-player-mpd-start-and-sync ()
   "Ensure that MusicPD's playlist is up-to-date with EMMS's
 playlist, and then play the current track.
 
 This is called if `emms-player-mpd-sync-playlist' is non-nil."
-  (emms-player-mpd-get-playlist-id
+  (when emms-player-mpd-status-timer
+    (emms-cancel-timer emms-player-mpd-status-timer)
+    (setq emms-player-mpd-status-timer nil))
+  (emms-player-mpd-send
+   "clearerror"
    nil
-   (lambda (closure id)
-     (let ((buf-id (with-current-emms-playlist emms-player-mpd-playlist-id)))
-       (if (and (not (buffer-modified-p emms-playlist-buffer))
-                (stringp buf-id)
-                (string= buf-id id))
-           (emms-player-mpd-start-and-sync-1 emms-playlist-buffer id)
-         (emms-player-mpd-sync-from-emms
-          #'emms-player-mpd-start-and-sync-1))))))
+   (lambda (closure response)
+     (emms-player-mpd-get-playlist-id
+      nil
+      #'emms-player-mpd-start-and-sync-1))))
 
 (defun emms-player-mpd-connect-1 (closure info)
   (setq emms-player-mpd-current-song nil)
@@ -812,7 +887,13 @@ This is called if `emms-player-mpd-sync-playlist' is non-nil."
 ;;;###autoload
 (defun emms-player-mpd-connect ()
   "Connect to MusicPD and retrieve its current playlist.
-Afterward, the status of MusicPD will be tracked."
+
+Afterward, the status of MusicPD will be tracked.
+
+This also has the effect of changing the current EMMS playlist to
+be the same as the current MusicPD playlist.  Thus, this
+function is useful to call if the contents of the EMMS playlist
+buffer get out-of-sync for some reason."
   (interactive)
   (when emms-player-mpd-status-timer
     (emms-cancel-timer emms-player-mpd-status-timer)
@@ -874,6 +955,17 @@ from other functions."
           (concat "seek " song " " (number-to-string (+ secs amount)))
           nil #'ignore))))))
 
+(defun emms-player-mpd-seek-to (pos)
+  "Seek to POS seconds from the start of the current track."
+  (interactive)
+  (emms-player-mpd-get-current-song
+   pos
+   (lambda (pos song)
+     (when (and song pos)
+       (emms-player-mpd-send
+        (concat "seek " song " " (number-to-string pos))
+        nil #'ignore)))))
+
 (defun emms-player-mpd-next ()
   "Move forward by one track in MusicPD's internal playlist."
   (interactive)
@@ -898,16 +990,6 @@ positive or negative."
         (concat "setvol \"" (number-to-string new-volume) "\"")
         nil #'ignore)))))
 
-(defun emms-volume-mpd-raise ()
-  "Increase the volume."
-  (interactive)
-  (emms-volume-mpd-change 5))
-
-(defun emms-volume-mpd-lower ()
-  "Decrease the volume."
-  (interactive)
-  (emms-volume-mpd-change -5))
-
 ;;; Now playing
 
 (defun emms-player-mpd-show-1 (closure response)
@@ -919,39 +1001,54 @@ positive or negative."
          (name (cdr (assoc "name" info))) ; radio feeds sometimes set this
          (file (cdr (assoc "file" info)))
          (desc nil))
-    (when info
-      (when name
-        (setq desc name))
-      (when file
-        (let ((track (emms-dictionary '*track*))
-              track-desc)
-          (if (string-match "\\`http://" file)
-              (emms-track-set track 'type 'url)
-            (emms-track-set track 'type 'file))
-          (emms-track-set track 'name file)
-          (emms-info-mpd track info)
-          (setq track-desc (emms-track-description track))
-          (when (and (stringp track-desc) (not (string= track-desc "")))
-            (setq desc (if desc
-                           (concat desc ": " track-desc)
-                         track-desc))))))
-    (if (not desc)
-        (message "Nothing playing right now")
-      (setq desc (format emms-show-format desc))
-      (cond ((functionp callback)
-             (funcall callback buffer desc))
-            (insertp
-             (when (buffer-live-p buffer)
-               (with-current-buffer buffer
-                 (insert desc))))
-            (t
-             (message "%s" desc))))))
+    ;; if we are playing lastfm radio, use its show function instead
+    (if (and (boundp 'emms-lastfm-radio-stream-url)
+             (stringp emms-lastfm-radio-stream-url)
+             (string= emms-lastfm-radio-stream-url file))
+        (with-current-buffer buffer
+          (and (fboundp 'emms-lastfm-np)
+               (emms-lastfm-np insertp callback)))
+      ;; otherwise build and show the description
+      (when info
+        (when name
+          (setq desc name))
+        (when file
+          (let ((track (emms-dictionary '*track*))
+                track-desc)
+            (if (string-match "\\`http://" file)
+                (emms-track-set track 'type 'url)
+              (emms-track-set track 'type 'file))
+            (emms-track-set track 'name file)
+            (emms-info-mpd track info)
+            (setq track-desc (emms-track-description track))
+            (when (and (stringp track-desc) (not (string= track-desc "")))
+              (setq desc (if desc
+                             (concat desc ": " track-desc)
+                           track-desc))))))
+      (if (not desc)
+          (unless (functionp callback)
+            (message "Nothing playing right now"))
+        (setq desc (format emms-show-format desc))
+        (cond ((functionp callback)
+               (funcall callback buffer desc))
+              (insertp
+               (when (buffer-live-p buffer)
+                 (with-current-buffer buffer
+                   (insert desc))))
+              (t
+               (message "%s" desc)))))))
 
 ;;;###autoload
 (defun emms-player-mpd-show (&optional insertp callback)
   "Describe the current EMMS track in the minibuffer.
-If INSERTP is non-nil, insert the description into the current buffer instead.
-If CALLBACK is a function, call it with the current buffer and description.
+
+If INSERTP is non-nil, insert the description into the current
+buffer instead.
+
+If CALLBACK is a function, call it with the current buffer and
+description as arguments instead of displaying the description or
+inserting it.
+
 This function uses `emms-show-format' to format the current track.
 It differs from `emms-show' in that it asks MusicPD for the current track,
 rather than EMMS."
@@ -988,9 +1085,10 @@ rather than EMMS."
 
 (defun emms-info-mpd (track &optional info)
   "Add track information to TRACK.
-This is a useful addition to `emms-info-functions'.
 If INFO is specified, use that instead of acquiring the necessary
-info from MusicPD."
+info from MusicPD.
+
+This is a useful addition to `emms-info-functions'."
   (if info
       (emms-info-mpd-process track info)
     (when (and (eq 'file (emms-track-type track))
@@ -1011,6 +1109,7 @@ info from MusicPD."
 
 (defun emms-cache-set-from-mpd-track (track-info)
   "Dump TRACK-INFO into the EMMS cache.
+
 The track should be an alist as per `emms-player-mpd-get-alist'."
   (when emms-cache-set-function
     (let ((track (emms-dictionary '*track*))
@@ -1024,6 +1123,7 @@ The track should be an alist as per `emms-player-mpd-get-alist'."
 
 (defun emms-cache-set-from-mpd-directory (dir)
   "Dump all MusicPD data from DIR into the EMMS cache.
+
 This is useful to do when you have recently acquired new music."
   (interactive
    (list (if emms-player-mpd-music-directory
@@ -1049,6 +1149,7 @@ This is useful to do when you have recently acquired new music."
 
 (defun emms-cache-set-from-mpd-all ()
   "Dump all MusicPD data into the EMMS cache.
+
 This is useful to do once, just before using emms-browser.el, in
 order to prime the cache."
   (interactive)
