@@ -1,24 +1,26 @@
 ;;; emms.el --- The Emacs Multimedia System
 
-;; Copyright (C) 2003, 2004, 2005, 2006 Free Software Foundation, Inc.
+;; Copyright (C) 2003, 2004, 2005, 2006, 2007 Free Software Foundation, Inc.
 
 ;; Author: Jorgen Schfer <forcer@forcix.cx>
 ;; Keywords: emms, mp3, mpeg, multimedia
 
-;; This file is free software; you can redistribute it and/or modify
+;; This file is part of EMMS.
+
+;; EMMS is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
-;; the Free Software Foundation; either version 2, or (at your option)
+;; the Free Software Foundation; either version 3, or (at your option)
 ;; any later version.
 ;;
-;; GNU Emacs is distributed in the hope that it will be useful,
+;; EMMS is distributed in the hope that it will be useful,
 ;; but WITHOUT ANY WARRANTY; without even the implied warranty of
 ;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ;; GNU General Public License for more details.
 ;;
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs; see the file COPYING. If not, write to the
+;; along with EMMS; see the file COPYING.  If not, write to the
 ;; Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
-;; Boston, MA 02110-1301 USA
+;; Boston, MA 02110-1301, USA.
 
 ;;; Commentary:
 
@@ -37,8 +39,7 @@
 
 ;;; Code:
 
-;; $Id: emms.el,v 1.63 2005/08/18 13:52:23 forcer Exp $
-(defvar emms-version "2.0"
+(defvar emms-version "3.0"
   "EMMS version string.")
 
 
@@ -109,6 +110,11 @@ songs, increase this number."
   :type 'function
   :group 'emms)
 
+(defcustom emms-playlist-uniq-function 'emms-playlist-simple-uniq
+  "*The function to use for make track uniq in the playlist."
+  :type 'function
+  :group 'emms)
+
 (defcustom emms-sort-lessp-function 'emms-sort-track-name-less-p
   "*Function for comparing two EMMS tracks.
 The function should return non-nil if and only if the first track
@@ -188,7 +194,7 @@ See `emms-player-finished-hook'."
   :group 'emms
   :type 'hook)
 
-(defcustom emms-player-finished-hook '(emms-next-noerror)
+(defcustom emms-player-finished-hook nil
   "*Hook run when an EMMS player finishes playing a track.
 Please pay attention to the differences between
 `emms-player-finished-hook' and `emms-player-stopped-hook'.
@@ -196,8 +202,14 @@ The former is called only when the player actually finishes
 playing a track; the latter, only when the player is stopped
 interactively."
   :group 'emms
-  :type 'hook
-  :options '(emms-next-noerror))
+  :type 'hook)
+
+(defcustom emms-player-next-function 'emms-next-noerror
+  "*A function run when EMMS thinks the next song should be played."
+  :group 'emms
+  :type 'function
+  :options '(emms-next-noerror
+             emms-random))
 
 (defcustom emms-player-paused-hook nil
   "*Hook run when a player is paused or resumed.
@@ -212,6 +224,32 @@ seconds the player did seek."
   :group 'emms
   :type 'hook)
 
+(defcustom emms-player-time-set-functions nil
+  "*Functions called when a player is setting the elapsed time of a track.
+The functions are called with a single argument, the time elapsed
+since the beginning of the current track."
+  :group 'emms
+  :type 'hook)
+
+(defcustom emms-cache-get-function nil
+  "A function to retrieve a track entry from the cache.
+This is called with two arguments, the type and the name."
+  :group 'emms
+  :type 'function)
+
+(defcustom emms-cache-set-function nil
+  "A function to add/set a track entry from the cache.
+This is called with three arguments: the type of the track, the
+name of the track, and the track itself."
+  :group 'emms
+  :type 'function)
+
+(defcustom emms-cache-modified-function nil
+  "A function to be called when a track is modified.
+The modified track is passed as the argument to this function."
+  :group 'emms
+  :type 'function)
+
 (defvar emms-player-playing-p nil
   "The currently playing EMMS player, or nil.")
 
@@ -223,6 +261,65 @@ seconds the player did seek."
 This can be used if the source depends on the current buffer not
 being the playlist buffer.")
 
+(defvar emms-playlist-buffer nil
+  "The current playlist buffer, if any.")
+
+
+;;; Macros
+
+;;; These need to be at the top of the file so that compilation works.
+
+(defmacro with-current-emms-playlist (&rest body)
+  "Run BODY with the current buffer being the current playlist buffer.
+This also disables any read-onliness of the current buffer."
+  `(progn
+     (when (or (not emms-playlist-buffer)
+               (not (buffer-live-p emms-playlist-buffer)))
+       (emms-playlist-current-clear))
+     (let ((emms-source-old-buffer (or emms-source-old-buffer
+                                       (current-buffer))))
+      (with-current-buffer emms-playlist-buffer
+        (let ((inhibit-read-only t))
+          ,@body)))))
+(put 'with-current-emms-playlist 'lisp-indent-function 0)
+(put 'with-current-emms-playlist 'edebug-form-spec '(body))
+
+(defmacro emms-with-inhibit-read-only-t (&rest body)
+  "Simple wrapper around `inhibit-read-only'."
+  `(let ((inhibit-read-only t))
+     ,@body))
+(put 'emms-with-inhibit-read-only-t 'edebug-form-spec '(body))
+
+(defmacro emms-with-widened-buffer (&rest body)
+  `(save-restriction
+     (widen)
+     ,@body))
+(put 'emms-with-widened-buffer 'edebug-form-spec '(body))
+
+(defmacro emms-walk-tracks (&rest body)
+  "Execute BODY for each track in the current buffer, starting at point.
+The point will be placed at the beginning of the track before
+executing BODY.
+
+The point will not be restored afterward."
+  (let ((donep (make-symbol "donep")))
+    `(let ((,donep nil))
+       ;; skip to first track if not on one
+       (unless (emms-playlist-track-at (point))
+         (condition-case nil
+             (emms-playlist-next)
+           (error
+            (setq ,donep t))))
+       ;; walk tracks
+       (while (not ,donep)
+         ,@body
+         (condition-case nil
+             (emms-playlist-next)
+           (error
+            (setq ,donep t)))))))
+(put 'emms-walk-tracks 'lisp-indent-function 0)
+(put 'emms-walk-tracks 'edebug-form-spec '(body))
+
 
 ;;; User Interface
 
@@ -240,7 +337,7 @@ being the playlist buffer.")
 
 (defun emms-next ()
   "Start playing the next track in the EMMS playlist.
-This might behave funny if called from `emms-player-finished-hook',
+This might behave funny if called from `emms-player-next-function',
 so use `emms-next-noerror' in that case."
   (interactive)
   (when emms-player-playing-p
@@ -253,7 +350,7 @@ so use `emms-next-noerror' in that case."
 Unlike `emms-next', this function doesn't signal an error when called
 at the end of the playlist.
 This function should only be called when no player is playing.
-This is a good function to put in `emms-player-finished-hook'."
+This is a good function to put in `emms-player-next-function'."
   (interactive)
   (when emms-player-playing-p
     (error "A track is already being played"))
@@ -298,6 +395,14 @@ It can also be negative to seek backwards."
   (emms-ensure-player-playing-p)
   (emms-player-seek seconds))
 
+(defun emms-seek-to (seconds)
+  "Seek the current player to SECONDS seconds.
+This can be a floating point number for sub-second fractions.
+It can also be negative to seek backwards."
+  (interactive "nSeconds to seek to: ")
+  (emms-ensure-player-playing-p)
+  (emms-player-seek-to seconds))
+
 (defun emms-seek-forward ()
   "Seek ten seconds forward."
   (interactive)
@@ -340,6 +445,14 @@ This uses `emms-playlist-sort-function'."
     (save-excursion
       (funcall emms-playlist-sort-function))))
 
+(defun emms-uniq ()
+  "Uniq the current playlist.
+This uses `emms-playlist-uniq-function'."
+  (interactive)
+  (with-current-emms-playlist
+    (save-excursion
+      (funcall emms-playlist-uniq-function))))
+
 (defun emms-toggle-repeat-playlist ()
   "Toggle whether emms repeats the playlist after it is done.
 See `emms-repeat-playlist'."
@@ -371,59 +484,37 @@ See  `emms-repeat-track'."
 
 ;;; Compatibility functions
 
-(defun emms-propertize (string &rest properties)
-  (if (fboundp 'propertize)
-      (apply #'propertize string properties)
-    (set-text-properties 0 (length string) properties string)
-    string))
+(require 'emms-compat)
 
-(defun emms-cancel-timer (timer)
-  "Cancel the given TIMER."
-  (when timer
-    (cond ((fboundp 'cancel-timer)
-           (cancel-timer timer))
-          ((fboundp 'delete-itimer)
-           (delete-itimer timer)))))
+
+;;; Dictionaries
 
-(defun emms-time-less-p (t1 t2)
-  "Say whether time T1 is less than time T2."
-  (or (< (car t1) (car t2))
-      (and (= (car t1) (car t2))
-           (< (nth 1 t1) (nth 1 t2)))))
+;; This is a simple helper data structure, used by both players
+;; and tracks.
 
-(defun emms-replace-regexp-in-string (regexp replacement text &optional fixedcase literal)
-  "Replace REGEXP with REPLACEMENT in TEXT.
-If fourth arg FIXEDCASE is non-nil, do not alter case of replacement text.
-If fifth arg LITERAL is non-nil, insert REPLACEMENT literally."
-  (cond
-   ((fboundp 'replace-regexp-in-string)
-    (replace-regexp-in-string regexp replacement text fixedcase literal))
-   ((fboundp 'replace-in-string)
-    (replace-in-string text regexp replacement literal))
-   (t (let ((repl-len (length replacement))
-            start)
-        (while (setq start (string-match regexp text start))
-          (setq start (+ start repl-len)
-                text (replace-match replacement fixedcase literal text))))
-      text)))
+(defsubst emms-dictionary (name)
+  "Create a new dictionary of type NAME."
+  (list name))
 
-(defun emms-line-number-at-pos (&optional pos)
-  "Return (narrowed) buffer line number at position POS.
-If POS is nil, use current buffer location."
-  (if (fboundp 'line-number-at-pos)
-      (line-number-at-pos pos)
-    (let ((opoint (or pos (point))) start)
-      (save-excursion
-        (goto-char (point-min))
-        (setq start (point))
-        (goto-char opoint)
-        (forward-line 0)
-        (1+ (count-lines start (point)))))))
+(defsubst emms-dictionary-type (dict)
+  "Return the type of the dictionary DICT."
+  (car dict))
 
-(defun emms-match-string-no-properties (num &optional string)
-  (if (fboundp 'match-string-no-properties)
-      (match-string-no-properties num string)
-    (match-string num string)))
+(defun emms-dictionary-get (dict name &optional default)
+  "Return the value of NAME in DICT."
+  (let ((item (assq name (cdr dict))))
+    (if item
+        (cdr item)
+      default)))
+
+(defun emms-dictionary-set (dict name value)
+  "Set the value of NAME in DICT to VALUE."
+  (let ((item (assq name (cdr dict))))
+    (if item
+        (setcdr item value)
+      (setcdr dict (append (cdr dict)
+                           (list (cons name value))))))
+  dict)
 
 
 ;;; Tracks
@@ -434,12 +525,26 @@ If POS is nil, use current buffer location."
 
 (defun emms-track (type name)
   "Create an EMMS track with type TYPE and name NAME."
-  (let ((track (emms-dictionary '*track*)))
-    (emms-track-set track 'type type)
-    (emms-track-set track 'name name)
+  (let ((track (when emms-cache-get-function
+                 (funcall emms-cache-get-function type name))))
+    (when (not track)
+      (setq track (emms-dictionary '*track*))
+      ;; Prevent the cache from being called for these two sets
+      (let ((emms-cache-modified-function nil))
+        (emms-track-set track 'type type)
+        (emms-track-set track 'name name))
+      (when emms-cache-set-function
+        (funcall emms-cache-set-function type name track)))
+    ;; run any hooks regardless of a cache hit, as the entry may be
+    ;; old
     (run-hook-with-args 'emms-track-initialize-functions track)
     track))
 
+(defun emms-track-p (obj)
+  "True if OBJ is an emms track."
+  (and (listp obj)
+       (eq (car obj) '*track*)))
+
 (defun emms-track-type (track)
   "Return the type of TRACK."
   (emms-track-get track 'type))
@@ -455,12 +560,18 @@ If there is no value, return DEFAULT (or nil, if not given)."
 
 (defun emms-track-set (track name value)
   "Set the value of NAME for TRACK to VALUE."
-  (emms-dictionary-set track name value))
+  (emms-dictionary-set track name value)
+  (when emms-cache-modified-function
+    (funcall emms-cache-modified-function track)))
 
 (defun emms-track-description (track)
   "Return a description of TRACK.
-This function uses `emms-track-description-function'."
-  (funcall emms-track-description-function track))
+This function uses the global value for `emms-track-description-function',
+rather than anything the current mode might have set.
+
+Use `emms-track-force-description' instead if you need to insert
+a description into a playlist buffer."
+  (funcall (default-value 'emms-track-description-function) track))
 
 (defun emms-track-updated (track)
   "Information in TRACK got updated."
@@ -474,8 +585,19 @@ Otherwise, return the type and the name with a colon in between."
   (if (eq 'file (emms-track-type track))
       (emms-track-name track)
     (concat (symbol-name (emms-track-type track))
-            ":"
-            (emms-track-name track))))
+            ": " (emms-track-name track))))
+
+(defun emms-track-force-description (track)
+  "Always return text that describes TRACK.
+This is used when inserting a description into a buffer.
+
+The reason for this is that if no text was returned (i.e. the
+user defined a track function that returned nil or the empty
+string), a confusing error message would result."
+  (let ((desc (funcall emms-track-description-function track)))
+    (if (and (stringp desc) (not (string= desc "")))
+        desc
+      (emms-track-simple-description track))))
 
 
 ;;; The Playlist
@@ -485,8 +607,11 @@ Otherwise, return the type and the name with a colon in between."
 ;; any kind of data. Strings of text with a `emms-track' property are
 ;; the tracks in the buffer.
 
-(defvar emms-playlist-buffer nil
-  "The current playlist buffer, if any.")
+(defvar emms-playlist-buffers nil
+  "The list of EMMS playlist buffers.
+You should use the `emms-playlist-buffer-list' function to
+retrieve a current list of EMMS buffers.  Never use this variable
+for that purpose.")
 
 (defvar emms-playlist-selected-marker nil
   "The marker for the currently selected track.")
@@ -497,34 +622,36 @@ Otherwise, return the type and the name with a colon in between."
 (make-variable-buffer-local 'emms-playlist-buffer-p)
 
 (defun emms-playlist-ensure-playlist-buffer ()
-  "Throw an error if we're on in a playlist-buffer."
+  "Throw an error if we're not in a playlist-buffer."
   (when (not emms-playlist-buffer-p)
     (error "Not an EMMS playlist buffer")))
 
-(defmacro with-current-emms-playlist (&rest body)
-  "Run BODY with the current buffer being the current playlist buffer.
-This also disables any read-onliness of the current buffer."
-  `(progn
-     (when (or (not emms-playlist-buffer)
-               (not (buffer-live-p emms-playlist-buffer)))
-       (emms-playlist-current-clear))
-     (let ((emms-source-old-buffer (or emms-source-old-buffer
-                                       (current-buffer))))
-      (with-current-buffer emms-playlist-buffer
-        (let ((inhibit-read-only t))
-          ,@body)))))
-(put 'with-current-emms-playlist 'lisp-indent-function 0)
-(put 'with-current-emms-playlist 'edebug-form-spec '(body))
-
 (defun emms-playlist-set-playlist-buffer (&optional buffer)
   "Set the current playlist buffer."
-  (interactive "bNew playlist buffer: ")
+  (interactive
+   (list (let* ((buf-list (mapcar #'(lambda (buf)
+                                      (list (buffer-name buf)))
+                                  (emms-playlist-buffer-list)))
+                (default (or (and emms-playlist-buffer-p
+                                  ;; default to current buffer
+                                  (buffer-name))
+                             ;; pick shortest buffer name, since it is
+                             ;; likely to be a shared prefix
+                             (car (sort buf-list
+                                        #'(lambda (lbuf rbuf)
+                                            (< (length (car lbuf))
+                                               (length (car rbuf)))))))))
+           (completing-read "Playlist buffer to make current: "
+                            buf-list nil t default))))
   (let ((buf (if buffer
                  (get-buffer buffer)
                (current-buffer))))
     (with-current-buffer buf
       (emms-playlist-ensure-playlist-buffer))
-    (setq emms-playlist-buffer buf)))
+    (setq emms-playlist-buffer buf)
+    (when (interactive-p)
+      (message "Set current EMMS playlist buffer"))
+    buf))
 
 (defun emms-playlist-new (&optional name)
   "Create a new playlist buffer.
@@ -538,10 +665,51 @@ If called interactively, the new buffer is also selected."
       (when (not (eq major-mode emms-playlist-default-major-mode))
         (funcall emms-playlist-default-major-mode))
       (setq emms-playlist-buffer-p t))
+    (add-to-list 'emms-playlist-buffers buf)
     (when (interactive-p)
       (switch-to-buffer buf))
     buf))
 
+(defun emms-playlist-buffer-list ()
+  "Return a list of EMMS playlist buffers.
+The first element is guaranteed to be the current EMMS playlist
+buffer, if it exists, otherwise the slot will be used for the
+other EMMS buffers.  The list will be in newest-first order."
+  ;; prune dead buffers
+  (setq emms-playlist-buffers (emms-delete-if (lambda (buf)
+                                                (not (buffer-live-p buf)))
+                                              emms-playlist-buffers))
+  ;; add new buffers
+  (mapc (lambda (buf)
+          (when (buffer-live-p buf)
+            (with-current-buffer buf
+              (when (and emms-playlist-buffer-p
+                         (not (memq buf emms-playlist-buffers)))
+                (setq emms-playlist-buffers
+                      (cons buf emms-playlist-buffers))))))
+        (buffer-list))
+  ;; force current playlist buffer to head position
+  (when (and (buffer-live-p emms-playlist-buffer)
+             (not (eq (car emms-playlist-buffers) emms-playlist-buffer)))
+    (setq emms-playlist-buffers (cons emms-playlist-buffer
+                                      (delete emms-playlist-buffer
+                                              emms-playlist-buffers))))
+  emms-playlist-buffers)
+
+(defun emms-playlist-current-kill ()
+  "Kill the current EMMS playlist buffer and switch to the next one."
+  (interactive)
+  (when (buffer-live-p emms-playlist-buffer)
+    (let ((new (cadr (emms-playlist-buffer-list))))
+      (if new
+          (let ((old emms-playlist-buffer))
+            (setq emms-playlist-buffer new
+                  emms-playlist-buffers (cdr emms-playlist-buffers))
+            (kill-buffer old)
+            (switch-to-buffer emms-playlist-buffer))
+        (with-current-buffer emms-playlist-buffer
+          (bury-buffer))))))
+
 (defun emms-playlist-current-clear ()
   "Clear the current playlist.
 If no current playlist exists, a new one is generated."
@@ -563,11 +731,6 @@ If no playlist exists, a new one is generated."
 		   (point-max)))
   (run-hooks 'emms-playlist-cleared-hook))
 
-(defmacro emms-with-widened-buffer (&rest body)
-  `(save-restriction
-     (widen)
-     ,@body))
-
 ;;; Point movement within the playlist buffer.
 (defun emms-playlist-track-at (&optional pos)
   "Return the track at POS (point if not given), or nil if none."
@@ -651,6 +814,17 @@ If no playlist exists, a new one is generated."
   (with-current-emms-playlist
     (emms-playlist-selected-track)))
 
+(defun emms-playlist-selected-track-at-p (&optional point)
+  "Return non-nil when POINT (defaulting to point) is on the selected track."
+  (when emms-playlist-selected-marker
+    (or (= emms-playlist-selected-marker
+           (or point (point)))
+        (let ((p (previous-single-property-change (or point (point))
+                                                  'emms-track)))
+          (when p
+            (= emms-playlist-selected-marker
+               p))))))
+
 (defun emms-playlist-select (pos)
   "Select the track at POS."
   (emms-playlist-ensure-playlist-buffer)
@@ -658,6 +832,7 @@ If no playlist exists, a new one is generated."
     (error "No track at position %s" pos))
   (when (not emms-playlist-selected-marker)
     (setq emms-playlist-selected-marker (make-marker)))
+  (set-marker-insertion-type emms-playlist-selected-marker t)
   (set-marker emms-playlist-selected-marker pos)
   (run-hooks 'emms-playlist-selection-changed-hook))
 
@@ -716,23 +891,11 @@ If no playlist exists, a new one is generated."
   (emms-playlist-ensure-playlist-buffer)
   ;; FIXME: This is rather inefficient.
   (save-excursion
-    (let ((track-indices nil)
-          (donep nil))
-      (condition-case nil
-          (progn
-            (emms-playlist-first)
-            (setq track-indices (cons (point)
-                                      track-indices)))
-        (error
-         (setq donep t)))
-      (while (not donep)
-        (condition-case nil
-            (progn
-              (emms-playlist-next)
-              (setq track-indices (cons (point)
-                                        track-indices)))
-          (error
-           (setq donep t))))
+    (let ((track-indices nil))
+      (goto-char (point-min))
+      (emms-walk-tracks
+        (setq track-indices (cons (point)
+                                  track-indices)))
       (setq track-indices (vconcat track-indices))
       (emms-playlist-select (aref track-indices
                                   (random (length track-indices)))))))
@@ -797,21 +960,13 @@ This is supplying ARGS as arguments to the source."
 (defun emms-playlist-tracks-in-region (beg end)
   "Return all tracks between BEG and END."
   (emms-playlist-ensure-playlist-buffer)
-  (let ((tracks nil)
-        (donep nil))
+  (let ((tracks nil))
     (save-restriction
       (narrow-to-region beg end)
-      (condition-case nil
-          (emms-playlist-first)
-        (error
-         (setq donep t)))
-      (while (not donep)
+      (goto-char (point-min))
+      (emms-walk-tracks
         (setq tracks (cons (emms-playlist-track-at (point))
-                           tracks))
-        (condition-case nil
-            (emms-playlist-next)
-          (error
-           (setq donep t)))))
+                           tracks))))
     tracks))
 
 (defun emms-playlist-track-updated (track)
@@ -839,7 +994,7 @@ This is supplying ARGS as arguments to the source."
   "Insert the description of TRACK at point."
   (emms-playlist-ensure-playlist-buffer)
   (let ((inhibit-read-only t))
-    (insert (emms-propertize (emms-track-description track)
+    (insert (emms-propertize (emms-track-force-description track)
                              'emms-track track)
             "\n")))
 
@@ -905,28 +1060,61 @@ ignore this."
           (emms-playlist-select pos)
         (emms-playlist-first)))))
 
+(defun emms-uniq-list (list stringify)
+  "Compare stringfied element of list, and remove duplicate elements."
+  ;; This uses a fast append list, keeping a pointer to the last cons
+  ;; cell of the list (TAIL). It might be worthwhile to provide an
+  ;; abstraction for this eventually.
+  (let* ((hash (make-hash-table :test 'equal))
+         (result (cons nil nil))
+         (tail result))
+    (dolist (element list)
+      (let ((str (funcall stringify element)))
+        (when (not (gethash str hash))
+          (setcdr tail (cons element nil))
+          (setq tail (cdr tail)))
+        (puthash str t hash)))
+    (cdr result)))
+
+(defun emms-playlist-simple-uniq ()
+  "Remove duplicate tracks"
+  (emms-playlist-ensure-playlist-buffer)
+  (widen)
+  (let ((inhibit-read-only t)
+        (current (emms-playlist-selected-track))
+        (tracks (emms-playlist-tracks-in-region (point-min)
+                                                (point-max))))
+    (delete-region (point-min) (point-max))
+    (run-hooks 'emms-playlist-cleared-hook)
+    (mapc 'emms-playlist-insert-track
+          (nreverse
+           (emms-uniq-list tracks 'emms-track-name)))
+    (let ((pos (text-property-any (point-min)
+                                  (point-max)
+                                  'emms-track current)))
+      (if pos
+          (emms-playlist-select pos)
+        (emms-playlist-first)))))
+
 ;;; Helper functions
 (defun emms-property-region (pos prop)
-  "Return a pair of the beginning and end of the property PROP at POS."
-  (let ((beg nil)
-        (end nil))
-    (save-excursion
-      (goto-char pos)
-      (while (and (not (bobp))
-                  (get-text-property (point)
-                                     prop))
-        (backward-char))
-      (when (not (get-text-property (point)
-                                    prop))
-        (forward-char))
-      (setq beg (point))
-      (goto-char pos)
-      (while (and (not (eobp))
-                  (get-text-property (point)
-                                     prop))
-        (forward-char))
-      (setq end (point)))
-    (cons beg end)))
+  "Return a pair of the beginning and end of the property PROP at POS.
+If POS does not contain PROP, try to find PROP just before POS."
+  (let (begin end)
+    (if (and (> pos (point-min))
+             (get-text-property (1- pos) prop))
+        (setq begin (previous-single-property-change (1- pos) prop))
+      (if (get-text-property pos prop)
+          (setq begin pos)
+        (error "Cannot find the %s property at the given position" prop)))
+    (if (get-text-property pos prop)
+        (setq end (next-single-property-change pos prop))
+      (if (and (> pos (point-min))
+               (get-text-property (1- pos) prop))
+          (setq end pos)
+        (error "Cannot find the %s property at the given position" prop)))
+    (cons (or begin (point-min))
+          (or end (point-max)))))
 
 (defun emms-shuffle-vector (vector)
   "Shuffle VECTOR."
@@ -981,11 +1169,17 @@ See emms-source-file.el for some examples."
        (defun ,source-play ,arglist
          ,docstring
          ,interactive
-         (emms-source-play ',source-name ,@call-args))
+         (if current-prefix-arg
+             (let ((current-prefix-arg nil))
+               (emms-source-add ',source-name ,@call-args))
+           (emms-source-play ',source-name ,@call-args)))
        (defun ,source-add ,arglist
          ,docstring
          ,interactive
-         (emms-source-add ',source-name ,@call-args))
+         (if current-prefix-arg
+             (let ((current-prefix-arg nil))
+               (emms-source-play ',source-name ,@call-args))
+           (emms-source-add ',source-name ,@call-args)))
        (defun ,source-insert ,arglist
          ,docstring
          ,interactive
@@ -1095,8 +1289,11 @@ or nil if no such player exists."
     (let ((player (emms-player-for track)))
       (if (not player)
           (error "Don't know how to play track: %S" track)
-        (funcall (emms-player-get player 'start)
-                 track)))))
+        ;; Change default-directory so we don't accidentally block any
+        ;; directories the current buffer was visiting.
+        (let ((default-directory "/"))
+          (funcall (emms-player-get player 'start)
+                   track))))))
 
 (defun emms-player-started (player)
   "Declare that the given EMMS PLAYER has started.
@@ -1119,7 +1316,8 @@ This should only be done by the current player itself."
   (if emms-player-stopped-p
       (run-hooks 'emms-player-stopped-hook)
     (sleep-for emms-player-delay)
-    (run-hooks 'emms-player-finished-hook)))
+    (run-hooks 'emms-player-finished-hook)
+    (funcall emms-player-next-function)))
 
 (defun emms-player-pause ()
   "Pause the current EMMS player."
@@ -1158,35 +1356,17 @@ or negative to seek backwards."
         (funcall seek seconds)
         (run-hook-with-args 'emms-player-seeked-functions seconds)))))
 
-
-;;; Dictionaries
-
-;; This is a simple helper data structure, used by both players
-;; and tracks.
-
-(defun emms-dictionary (name)
-  "Create a new dictionary of type NAME."
-  (list name))
-
-(defun emms-dictionary-type (dict)
-  "Return the type of the dictionary DICT."
-  (car dict))
-
-(defun emms-dictionary-get (dict name &optional default)
-  "Return the value of NAME in DICT."
-  (let ((item (assq name (cdr dict))))
-    (if item
-        (cdr item)
-      default)))
-
-(defun emms-dictionary-set (dict name value)
-  "Set the value of NAME in DICT to VALUE."
-  (let ((item (assq name (cdr dict))))
-    (if item
-        (setcdr item value)
-      (setcdr dict (append (cdr dict)
-                           (list (cons name value))))))
-  dict)
+(defun emms-player-seek-to (seconds)
+  "Seek the current player to SECONDS seconds.
+This can be a floating point number for fractions of a second,
+or negative to seek backwards."
+  (if (not emms-player-playing-p)
+      (error "Can't seek-to player, nothing playing right now")
+    (let ((seek (emms-player-get emms-player-playing-p 'seek-to)))
+      (if (not seek)
+          (error "Player does not know how to seek-to")
+        (funcall seek seconds)
+        (run-hook-with-args 'emms-player-time-set-functions seconds)))))
 
 (provide 'emms)
 ;;; emms.el ends here
